@@ -1,11 +1,11 @@
-var CodeRef = require('./code-ref.js');
+const CodeRef = require('./code-ref.js');
 class CodeRefWithStateVars extends CodeRef {
   freshBlock() {
     return this.closureTemplate();
   }
 
   $$clone() {
-    var clone = new CodeRefWithStateVars(this.name, undefined);
+    const clone = new CodeRefWithStateVars(this.name, undefined);
     clone.$$call = this.freshBlock();
     clone.codeObj = this.codeObj;
     clone.staticCode = this.staticCode;
@@ -20,7 +20,7 @@ class CodeRefWithStateVars extends CodeRef {
   }
 
   closureCtx(outerCtx) {
-    var closure = new CodeRefWithStateVars(this.name, this.cuid);
+    const closure = new CodeRefWithStateVars(this.name, this.cuid);
     closure.codeObj = this.codeObj;
     closure.$$call = this.freshBlock();
     closure.closureTemplate = this.closureTemplate;
@@ -30,5 +30,17 @@ class CodeRefWithStateVars extends CodeRef {
 
     return closure;
   }
+
+  captureAndClosureCtx(outerCtx) {
+    this.capture(this.freshBlock());
+    const closure = this.closureCtx(outerCtx);
+
+    if (outerCtx !== null) {
+      this.outerCtx = outerCtx;
+      closure.outerCtx = outerCtx;
+    }
+
+    return closure;
+  }
 };
 module.exports = CodeRefWithStateVars;
