@@ -1,7 +1,7 @@
 'use strict';
 
-var NQPObject = require('./nqp-object.js');
-var StaticCtx = require('./static-ctx.js');
+const NQPObject = require('./nqp-object.js');
+const StaticCtx = require('./static-ctx.js');
 
 class CodeRef extends NQPObject {
   freshBlock() {
@@ -32,7 +32,7 @@ class CodeRef extends NQPObject {
 
   // HACK - do this properly
   $$call() {
-    let staticCode = this.staticCode;
+    const staticCode = this.staticCode;
     if (staticCode.closureTemplate) {
       /* HACK */
       let searched = staticCode;
@@ -40,7 +40,7 @@ class CodeRef extends NQPObject {
       let fakeCtx;
       let fakeOuterCtx;
 
-      let setOuter = outer => {
+      const setOuter = outer => {
         if (fakeCtx) {
           fakeCtx.$$outer = outer;
         } else {
@@ -53,20 +53,28 @@ class CodeRef extends NQPObject {
           setOuter(searched.outerCtx);
           break;
         } else if (searched.outerCodeRef) {
-          let newFakeCtx = new StaticCtx();
+          const newFakeCtx = new StaticCtx();
 
           setOuter(newFakeCtx);
 
           fakeCtx = newFakeCtx;
 
-          let outerCode = searched.outerCodeRef;
+          const outerCode = searched.outerCodeRef;
           fakeCtx.$$cuid = searched.outerCodeRef.cuid;
-          let staticVars = outerCode.staticVars;
+          const staticVars = outerCode.staticVars;
           if (staticVars) {
-            for (var staticVarName in staticVars) {
+            for (const staticVarName in staticVars) {
               fakeCtx[staticVarName] = staticVars[staticVarName];
             }
           }
+
+          const contVars = outerCode.contVars;
+          if (contVars) {
+            for (const contVarName in contVars) {
+              /* not cloning make INIT and BEGIN correctly */
+              fakeCtx[contVarName] = contVars[contVarName];
+            }
+          }
           searched = searched.outerCodeRef;
         } else {
           setOuter(null);
@@ -77,7 +85,7 @@ class CodeRef extends NQPObject {
       this.outerCtx = fakeOuterCtx;
 
       this.$$call = staticCode.freshBlock();
-      return staticCode.$$call.apply(staticCode, arguments);
+      return this.$$call.apply(this, arguments);
     } else {
       console.log('can\'t autoclose - BAD');
     }
@@ -88,7 +96,7 @@ class CodeRef extends NQPObject {
   }
 
   closure(block) {
-    var closure = new CodeRef(this.name, this.cuid);
+    const closure = new CodeRef(this.name, this.cuid);
     closure.codeObj = this.codeObj;
     closure.$$call = block;
     closure.staticCode = this;
@@ -96,7 +104,7 @@ class CodeRef extends NQPObject {
   }
 
   closureCtx(outerCtx) {
-    var closure = new CodeRef(this.name, this.cuid);
+    const closure = new CodeRef(this.name, this.cuid);
     closure.codeObj = this.codeObj;
     closure.$$call = this.freshBlock();
     closure.staticCode = this;
@@ -115,7 +123,7 @@ class CodeRef extends NQPObject {
 
   captureAndClosure(outer, block) {
     this.capture(block);
-    var closure = this.closure(block);
+    const closure = this.closure(block);
     if (outer !== null) {
       this.outerCtx = outer;
       closure.outerCtx = outer;
@@ -125,7 +133,7 @@ class CodeRef extends NQPObject {
 
   captureAndClosureCtx(outerCtx) {
     this.capture(this.freshBlock());
-    var closure = this.closure(this.freshBlock());
+    const closure = this.closure(this.freshBlock());
 
     if (outerCtx !== null) {
       this.outerCtx = outerCtx;
@@ -140,16 +148,18 @@ class CodeRef extends NQPObject {
     return this;
   }
 
-  setInfo(outerCodeRef, closureTemplate, lexicalsTypeInfo, staticVars) {
+  setInfo(outerCodeRef, closureTemplate, lexicalsTypeInfo, staticVars, contVars, sourceOffset) {
     this.closureTemplate = closureTemplate;
     this.outerCodeRef = outerCodeRef;
     this.lexicalsTypeInfo = lexicalsTypeInfo;
     this.staticVars = staticVars;
+    this.contVars = contVars;
+    this.sourceOffset = sourceOffset;
     return this;
   }
 
   $$clone() {
-    var clone = new CodeRef(this.name, undefined);
+    const clone = new CodeRef(this.name, undefined);
     clone.$$call = this.$$call;
     clone.codeObj = this.codeObj;
     clone.staticCode = this.staticCode;
@@ -168,6 +178,10 @@ class CodeRef extends NQPObject {
   $$istype(ctx, type) {
     return 0;
   }
+
+  getOuterCtx() {
+    return this.outerCtx || this.staticCode.outerCtx || null;
+  }
 };
 
 module.exports = CodeRef;
