@@ -1,8 +1,7 @@
 'use strict';
-var Ctx = require('./ctx.js');
+const Ctx = require('./ctx.js');
 
 class CtxJustHandler extends Ctx {
-
   $$skipHandlers() {
     return this.$$caller.$$skipHandlers();
   }
