@@ -1,4 +1,6 @@
-let NQPObject = require('./nqp-object.js');
+const NQPObject = require('./nqp-object.js');
+const NQPStr = require('./nqp-str.js');
+
 class HashIter extends NQPObject {
   constructor(hash) {
     super();
@@ -25,15 +27,15 @@ class HashIter extends NQPObject {
     return this.$$idx < this.$$target;
   }
 
-  Str(ctx, _NAMED, self) {
-    return this.$$iterkey_s();
+  p6$Str(ctx, _NAMED, self) {
+    return new NQPStr(this.$$iterkey_s());
   }
 
-  key(ctx, _NAMED, self) {
-    return this.$$iterkey_s();
+  p6$key(ctx, _NAMED, self) {
+    return new NQPStr(this.$$iterkey_s());
   }
 
-  value(ctx, _NAMED, self) {
+  p6$value(ctx, _NAMED, self) {
     return this.$$iterval();
   }
 };
