@@ -1,6 +1,8 @@
 'use strict';
 
-var NQPObject = require('./nqp-object.js');
+const globalContext = require('./global-context.js');
+
+const NQPObject = require('./nqp-object.js');
 
 /**
  * A serialization context holds a list of objects and code references that live
@@ -25,7 +27,6 @@ class SerializationContext extends NQPObject {
     super();
     this.description = '???';
     this.handle = handle;
-    SerializationContext.contexts[handle] = this;
     this.rootObjects = [];
     this.rootSTables = [];
     this.rootCodes = [];
@@ -42,9 +43,9 @@ class SerializationContext extends NQPObject {
 
   setObj(idx, obj) {
     this.rootObjects[idx] = obj;
-    if (!obj._STable._SC) {
-      this.rootSTables.push(obj._STable);
-      obj._STable._SC = this;
+    if (!obj.$$STable.$$SC) {
+      this.rootSTables.push(obj.$$STable);
+      obj.$$STable.$$SC = this;
     }
   }
 
@@ -53,23 +54,23 @@ class SerializationContext extends NQPObject {
   }
 
   repossessSTable(STable) {
-    var newSlot = this.rootSTables.length;
+    const newSlot = this.rootSTables.length;
     this.rootSTables.push(STable);
 
     this.repIndexes.push((newSlot << 1) | 1);
-    this.repScs.push(STable._SC);
+    this.repScs.push(STable.$$SC);
 
-    STable._SC = this;
+    STable.$$SC = this;
   }
 
   repossessObject(obj) {
-    var newSlot = this.rootObjects.length;
+    const newSlot = this.rootObjects.length;
     this.rootObjects.push(obj);
 
     this.repIndexes.push((newSlot << 1));
-    this.repScs.push(obj._SC);
+    this.repScs.push(obj.$$SC);
 
-    obj._SC = this;
+    obj.$$SC = this;
   }
 
   $$toBool(ctx) {
@@ -77,5 +78,4 @@ class SerializationContext extends NQPObject {
   }
 };
 
-SerializationContext.contexts = {};
 module.exports = SerializationContext;
