@@ -1,34 +1,56 @@
 'use strict';
 
-var nullStr = require('./null_s.js'); /* Used when evaling runtime compiled methods */
-var Null = require('./null.js');
+const JSBI = require('jsbi');
+const ZERO = JSBI.BigInt(0);
 
-var repossession = require('./repossession.js');
-var compilingSCs = repossession.compilingSCs;
+/* Used when evaling runtime compiled methods */
+const nullStr = require('./null_s.js');
 
-var constants = require('./constants.js');
+const Null = require('./null.js');
+
+const globalContext = require('./global-context.js');
+
+const constants = require('./constants.js');
 
 /* Needed for setting defaults values of attrs for objects */
-var bignum = require('bignum-browserify');
-var ZERO = bignum(0);
 
-function findMethod(ctx, obj, name) {
-  if (obj._STable.methodCache) {
-    let method = obj._STable.methodCache.get(name);
+/* Needed for throwing exceptions from generated code */
+const NQPException = require('./nqp-exception.js');
+
+const nativeArgs = require('./native-args.js');
+const NativeStrArg = nativeArgs.NativeStrArg;
+
+/*async*/ function findMethod(ctx, obj, name) {
+  if (obj.$$STable.methodCache) {
+    const method = obj.$$STable.methodCache.get(name);
     if (method !== undefined) {
       return method;
     }
-    if (obj._STable.modeFlags & constants.METHOD_CACHE_AUTHORITATIVE) {
+    if (obj.$$STable.modeFlags & constants.METHOD_CACHE_AUTHORITATIVE) {
       return Null;
     }
   }
 
-  if (obj._STable.HOW.$$can(ctx, 'find_method')) {
-    return obj._STable.HOW.find_method(ctx, null, obj._STable.HOW, obj, name);
+  if (/*await*/ obj.$$STable.HOW.$$can(ctx, 'find_method')) {
+    return /*await*/ obj.$$STable.HOW.p6$find_method(ctx, null, obj.$$STable.HOW, obj, new NativeStrArg(name));
   } else {
     return Null;
   }
 }
+function scwb() {
+  const compilingSCs = globalContext.context.compilingSCs;
+
+  if (compilingSCs.length == 0 || globalContext.context.scwbDisableDepth || globalContext.context.neverRepossess.has(this)) {
+    return;
+  }
+
+  if (compilingSCs[compilingSCs.length - 1] !== this.$$SC) {
+    const owned = this.$$SC.ownedObjects.get(this);
+    compilingSCs[compilingSCs.length - 1].repossessObject(owned === undefined ? this : owned);
+  }
+}
+
+module.exports.scwb = scwb;
 
 class STable {
   constructor(REPR, HOW) {
@@ -39,15 +61,15 @@ class STable {
 
     this.lazyMethodCache = false;
 
-    this._SC = undefined;
+    this.$$SC = undefined;
 
     this.ObjConstructor = REPR.createObjConstructor(this);
 
     /* HACK - it's a bit hackish - think how correct it is */
     this.ObjConstructor.prototype.$$clone = function() {
-      var clone = new this._STable.ObjConstructor();
-      for (var i in this) {
-        if (Object.prototype.hasOwnProperty.call(this, i) && i != '_SC') {
+      const clone = new this.$$STable.ObjConstructor();
+      for (const i in this) {
+        if (Object.prototype.hasOwnProperty.call(this, i) && i != '$$SC') {
           clone[i] = this[i];
         }
       }
@@ -56,7 +78,7 @@ class STable {
 
     /* Default boolification mode 5 */
     this.ObjConstructor.prototype.$$toBool = function(ctx) {
-      return this.typeObject_ ? 0 : 1;
+      return this.$$typeObject ? 0 : 1;
     };
 
     this.ObjConstructor.prototype.$$decont = function(ctx) {
@@ -67,25 +89,15 @@ class STable {
       return 0;
     };
 
-    this.ObjConstructor.prototype.typeObject_ = 0;
+    this.ObjConstructor.prototype.$$typeObject = 0;
 
     this.ObjConstructor.prototype.$$call = undefined;
 
-    this.ObjConstructor.prototype.$$scwb = function() {
-      if (compilingSCs.length == 0 || repossession.scwbDisableDepth || repossession.neverRepossess.get(this)) {
-        return;
-      }
-
-      if (compilingSCs[compilingSCs.length - 1] !== this._SC) {
-        var owned = this._SC.ownedObjects.get(this);
-        compilingSCs[compilingSCs.length - 1].repossessObject(owned === undefined ? this : owned);
-      }
-    };
-
-    this.ObjConstructor.prototype.$$istype = function(ctx, type) {
-      var cache = this._STable.typeCheckCache;
+    this.ObjConstructor.prototype.$$scwb = scwb;
+    this.ObjConstructor.prototype.$$istype = /*async*/ function(ctx, type) {
+      const cache = this.$$STable.typeCheckCache;
       if (cache) {
-        for (var i = 0; i < cache.length; i++) {
+        for (let i = 0; i < cache.length; i++) {
           if (cache[i] === type) {
             return 1;
           }
@@ -94,27 +106,28 @@ class STable {
         /* If we get here, need to call .^type_check on the value we're
          * checking. */
 
-        var HOW = this._STable.HOW;
+        const HOW = this.$$STable.HOW;
         /* This "hack" is stolen from the JVM */
-        if (!HOW.$$can(ctx, 'type_check')) {
+        if (!/*await*/ HOW.$$can(ctx, 'type_check')) {
           return 0;
         }
 
-        if (HOW.type_check(ctx, null, HOW, this, type).$$toBool(ctx)) {
+        const typeCheckResult = /*await*/ HOW.p6$type_check(ctx, null, HOW, this, type);
+        if (typeof typeCheckResult === 'number' ? typeCheckResult : typeCheckResult.$$toBool(ctx)) {
           return 1;
         }
       }
 
       const TYPE_CHECK_NEEDS_ACCEPTS = 2;
-      if (type._STable.modeFlags & TYPE_CHECK_NEEDS_ACCEPTS) {
-        return type._STable.HOW.accepts_type(ctx, null, type._STable.HOW, type, this).$$toBool(ctx);
+      if (type.$$STable.modeFlags & TYPE_CHECK_NEEDS_ACCEPTS) {
+        return (/*await*/ type.$$STable.HOW.p6$accepts_type(ctx, null, type.$$STable.HOW, type, this)).$$toBool(ctx);
       }
 
       return 0;
     };
 
-    this.ObjConstructor.prototype.$$can = function(ctx, name) {
-      return findMethod(ctx, this, name) === Null ? 0 : 1;
+    this.ObjConstructor.prototype.$$can = /*async*/ function(ctx, name) {
+      return (/*await*/ findMethod(ctx, this, name)) === Null ? 0 : 1;
     };
 
     if (this.REPR.setupSTable) {
@@ -122,34 +135,56 @@ class STable {
     }
   }
 
+  lookupParametric(params) {
+    const lookup = this.parameterizerCache;
+    for (let i = 0; i < lookup.length; i++) {
+      if (params.length == lookup[i].params.length) {
+        let match = true;
+        for (let j = 0; j < params.length; j++) {
+          /* XXX More cases to consider here. - copied over from the jvm backend, need to consider what they are*/
+          if (params[j] !== lookup[i].params[j]) {
+            match = false;
+            break;
+          }
+        }
+
+        if (match) {
+          return lookup[i].type;
+        }
+      }
+    }
+  }
+
   setboolspec(mode, method) {
     this.boolificationSpec = {mode: mode, method: method};
     if (mode == 0) {
-      this.ObjConstructor.prototype.$$toBool = function(ctx) {
-        return method.$$call(ctx, {}, this).$$decont().$$toBool(ctx);
+      this.ObjConstructor.prototype.$$toBool = /*async*/ function(ctx) {
+        const ret = /*await*/ method.$$call(ctx, {}, this);
+        return (typeof ret === 'number' ? (ret === 0 ? 0 : 1) : ret.$$decont().$$toBool(ctx));
       };
     } else if (mode == 1) {
       this.ObjConstructor.prototype.$$toBool = function(ctx) {
-        return this.typeObject_ || this.$$getInt() == 0 ? 0 : 1;
+        return this.$$typeObject || this.$$getInt() == 0 ? 0 : 1;
       };
     } else if (mode == 2) {
       this.ObjConstructor.prototype.$$toBool = function(ctx) {
-        return this.typeObject_ || this.$$getNum() == 0 ? 0 : 1;
+        return this.$$typeObject || this.$$getNum() == 0 ? 0 : 1;
       };
     } else if (mode == 3) {
       this.ObjConstructor.prototype.$$toBool = function(ctx) {
-        return this.typeObject_ || this.$$getStr() == '' ? 0 : 1;
+        return (this.$$typeObject || this.$$getStr() === nullStr || this.$$getStr() === '')  ? 0 : 1;
       };
     } else if (mode == 4) {
       this.ObjConstructor.prototype.$$toBool = function(ctx) {
-        var str = this.$$getStr();
-        return this.typeObject_ || (str == '' || str == '0') ? 0 : 1;
+        if (this.$$typeObject) return 0;
+        const str = this.$$getStr();
+        return (str === nullStr || str == '' || str == '0') ? 0 : 1;
       };
     } else if (mode == 5) {
     // this is the default - do nothing
     } else if (mode == 6) {
       this.ObjConstructor.prototype.$$toBool = function(ctx) {
-        return (this.typeObject_ || this.$$getBignum().eq(0)) ? 0 : 1;
+        return (this.$$typeObject || JSBI.equal(this.$$getBignum(), ZERO)) ? 0 : 1;
       };
     } else if (mode == 7) {
     // STUB
@@ -158,16 +193,20 @@ class STable {
         return this.$$elems() ? 1 : 0;
       };
     } else {
-      throw 'setboolspec with mode: ' + mode + ' NYI';
+      throw new NQPException('setboolspec with mode: ' + mode + ' NYI');
     }
   }
 
   setinvokespec(classHandle, attrName, invocationHandler) {
+    if (this.REPR.hardcodedInvokeSpec) {
+      return;
+    }
+
     if (classHandle !== Null) {
       /* TODO  - think if we can use direct access here */
-      var getter = this.REPR.getterForAttr(classHandle, attrName);
+      const getter = this.REPR.getterForAttr(classHandle, attrName);
       this.ObjConstructor.prototype.$$call = function() {
-        var value = this[getter]();
+        const value = this[getter]();
         return value.$$call.apply(value, arguments);
       };
       this.ObjConstructor.prototype.$$apply = function(args) {
@@ -175,22 +214,22 @@ class STable {
       };
     } else {
       this.ObjConstructor.prototype.$$call = function() {
-        var args = [];
+        const args = [];
         args.push(arguments[0]);
         args.push(arguments[1]);
         args.push(this);
-        for (var i = 2; i < arguments.length; i++) {
+        for (let i = 2; i < arguments.length; i++) {
           args.push(arguments[i]);
         }
         return invocationHandler.$$apply(args);
       };
 
       this.ObjConstructor.prototype.$$apply = function(args) {
-        var newArgs = [];
+        const newArgs = [];
         newArgs.push(args[0]);
         newArgs.push(args[1]);
         newArgs.push(this);
-        for (var i = 2; i < args.length; i++) {
+        for (let i = 2; i < args.length; i++) {
           newArgs.push(args[i]);
         }
         return invocationHandler.$$apply(newArgs);
@@ -201,8 +240,8 @@ class STable {
 
 
   createTypeObject() {
-    var obj = new this.ObjConstructor();
-    obj.typeObject_ = 1;
+    const obj = new this.ObjConstructor();
+    obj.$$typeObject = 1;
     obj.$$atkey = function(key) {
       return Null;
     };
@@ -212,6 +251,30 @@ class STable {
     obj.$$decont = function(ctx) {
       return this;
     };
+    obj.$$clone = function(ctx) {
+      return this;
+    };
+
+    obj.$$getInt = function() {
+      throw new NQPException(`Cannot unbox a type object (${this.$$STable.debugName}) to an int.`);
+    };
+
+    obj.$$getNum = function() {
+      throw new NQPException(`Cannot unbox a type object (${this.$$STable.debugName}) to an num.`);
+    };
+
+    obj.$$getStr = function() {
+      throw new NQPException(`Cannot unbox a type object (${this.$$STable.debugName}) to an str.`);
+    };
+
+    obj.$$getattr = function(classHandle, attrName) {
+      throw new NQPException(`Cannot look up attributes in a ${this.$$STable.debugName} type object`);
+    };
+
+    obj.$$bindattr = function(classHandle, attrName, value) {
+      throw new NQPException(`Cannot bind attributes in a ${this.$$STable.debugName} type object`);
+    };
+
     return obj;
   }
 
@@ -227,10 +290,10 @@ class STable {
 
     this.lazyMethodCache = false;
 
-    let proto = this.ObjConstructor.prototype;
+    const proto = this.ObjConstructor.prototype;
 
     methodCache.forEach(function(method, name, map) {
-      proto[name] = function() {
+      proto['p6$' + name] = function() {
         return method.$$call.apply(method, arguments);
       };
     });
@@ -286,11 +349,15 @@ class STable {
   }
 
   addInternalMethods(klass) {
-    for (let methodName of Object.getOwnPropertyNames(klass.prototype)) {
+    for (const methodName of Object.getOwnPropertyNames(klass.prototype)) {
       this.addInternalMethod(methodName, klass.prototype[methodName]);
     }
   }
 
+  addCustomInspection(callback) {
+    this.addInternalMethod(Symbol.for('nodejs.util.inspect.custom'), callback);
+  }
+
   compileAccessor(accessor, code, setup) {
     this.code = this.code || '';
     this.setup = this.setup || '';
@@ -300,15 +367,17 @@ class STable {
 
   evalGatheredCode() {
     if (this.code) {
-      var STable = this; // eslint-disable-line no-unused-vars
+      const STable = this; // eslint-disable-line no-unused-vars
       eval(this.setup + this.code);
       this.code = '';
     }
   }
 
   scwb() {
-    if (compilingSCs.length == 0 || repossession.scwbDisableDepth) return;
-    if (compilingSCs[compilingSCs.length - 1] !== this._SC) {
+    const compilingSCs = globalContext.context.compilingSCs;
+
+    if (compilingSCs.length == 0 || globalContext.context.scwbDisableDepth) return;
+    if (compilingSCs[compilingSCs.length - 1] !== this.$$SC) {
       compilingSCs[compilingSCs.length - 1].repossessSTable(this);
     }
   }
